#include "AdaptInstationary.hpp"

// AMDiS includes
#include <amdis/AdaptInfo.hpp>
#include <amdis/Flag.hpp>
#include <amdis/Initfile.hpp>
#include <amdis/Output.hpp>
#include <amdis/ProblemIterationInterface.hpp>
#include <amdis/ProblemTimeInterface.hpp>

namespace AMDiS
{
  AdaptInstationary::AdaptInstationary(std::string name,
                                       ProblemIterationInterface& problemStat,
                                       AdaptInfo& adaptInfo,
                                       ProblemTimeInterface& problemInstat,
                                       AdaptInfo& initialInfo)
    : AdaptBase(name, &problemStat, adaptInfo, &problemInstat, &initialInfo)
  {
    Parameters::get(name + "->strategy", strategy);
    Parameters::get(name + "->time delta 1", timeDelta1);
    Parameters::get(name + "->time delta 2", timeDelta2);
    Parameters::get(name + "->info", info);
    Parameters::get(name + "->break when stable", breakWhenStable);

    fixedTimestep = (adaptInfo.getMinTimestep() == adaptInfo.getMaxTimestep());
  }


  void AdaptInstationary::explicitTimeStrategy()
  {
    AMDIS_FUNCNAME("AdaptInstationary::explicitTimeStrategy()");

    // estimate before first adaption
    if (adaptInfo.getTime() <= adaptInfo.getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);


    // increment time
    adaptInfo.setTime(adaptInfo.getTime() + adaptInfo.getTimestep());

    problemTime->setTime(adaptInfo);

    msg("time = {}, timestep = {}", adaptInfo.getTime(), adaptInfo.getTimestep());

    adaptInfo.setSpaceIteration(0);

    // do the iteration
    problemIteration->beginIteration(adaptInfo);
    problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
    problemIteration->endIteration(adaptInfo);
    adaptInfo.setLastProcessedTimestep(adaptInfo.getTimestep());
  }


  void AdaptInstationary::implicitTimeStrategy()
  {
    AMDIS_FUNCNAME("AdaptInstationary::implicitTimeStrategy()");

    do
    {
      adaptInfo.setTime(adaptInfo.getTime() + adaptInfo.getTimestep());
      problemTime->setTime(adaptInfo);

      msg("time = {}, timestep = {}", adaptInfo.getTime(), adaptInfo.getTimestep());

      problemIteration->oneIteration(adaptInfo, NO_ADAPTION);

      adaptInfo.incTimestepIteration();

      if (!fixedTimestep &&
          !adaptInfo.timeToleranceReached() &&
          adaptInfo.getTimestepIteration() <= adaptInfo.getMaxTimestepIteration() &&
          !(adaptInfo.getTimestep() <= adaptInfo.getMinTimestep()))
      {
        adaptInfo.setTime(adaptInfo.getTime() - adaptInfo.getTimestep());
        adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta1);
        continue;
      }


      adaptInfo.setSpaceIteration(0);


      // === Do space iterations only if the maximum is higher than 0. ===

      if (adaptInfo.getMaxSpaceIteration() > 0)
      {

        // === Space iterations. ===
        do
        {
          problemIteration->beginIteration(adaptInfo);

          Flag adapted = problemIteration->oneIteration(adaptInfo, FULL_ITERATION);
          if (adapted == Flag{0})
          {
            if (!fixedTimestep &&
                !adaptInfo.timeToleranceReached() &&
                !(adaptInfo.getTimestep() <= adaptInfo.getMinTimestep()))
            {
              adaptInfo.setTime(adaptInfo.getTime() - adaptInfo.getTimestep());
              adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta2);
              problemIteration->endIteration(adaptInfo);
              adaptInfo.incSpaceIteration();
              break;
            }
          }

          adaptInfo.incSpaceIteration();
          problemIteration->endIteration(adaptInfo);

        }
        while (!adaptInfo.spaceToleranceReached() &&
               adaptInfo.getSpaceIteration() <= adaptInfo.getMaxSpaceIteration());

      }
      else
      {
        problemIteration->endIteration(adaptInfo);
      }


    }
    while(!adaptInfo.timeToleranceReached() &&
          !(adaptInfo.getTimestep() <= adaptInfo.getMinTimestep()) &&
          adaptInfo.getTimestepIteration() <= adaptInfo.getMaxTimestepIteration());

    adaptInfo.setLastProcessedTimestep(adaptInfo.getTimestep());

    // After successful iteration/timestep the timestep will be changed according
    // adaption rules for next timestep.
    // First, check for increase of timestep
    if (!fixedTimestep && adaptInfo.timeErrorLow())
      adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta2);

    // Second, check for decrease of timestep
    if (!fixedTimestep &&
        !adaptInfo.timeToleranceReached() &&
        !(adaptInfo.getTimestep() <= adaptInfo.getMinTimestep()))
      adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta1);
  }


  void AdaptInstationary::simpleAdaptiveTimeStrategy()
  {
    AMDIS_FUNCNAME("AdaptInstationary::simpleAdaptiveTimeStrategy()");

    // estimate before first adaption
    if (adaptInfo.getTime() <= adaptInfo.getStartTime())
      problemIteration->oneIteration(adaptInfo, ESTIMATE);

    adaptInfo.setTime(adaptInfo.getTime() + adaptInfo.getTimestep());
    problemTime->setTime(adaptInfo);

    msg("time = {}, timestep = {}", adaptInfo.getTime(), adaptInfo.getTimestep());

    problemIteration->oneIteration(adaptInfo, FULL_ITERATION);

    adaptInfo.setLastProcessedTimestep(adaptInfo.getTimestep());

    // First, check for increase of timestep
    if (!fixedTimestep && adaptInfo.timeErrorLow())
      adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta2);

    // Second, check for decrease of timestep
    if (!fixedTimestep &&
        !adaptInfo.timeToleranceReached() &&
        !(adaptInfo.getTimestep() <= adaptInfo.getMinTimestep()))
      adaptInfo.setTimestep(adaptInfo.getTimestep() * timeDelta1);
  }


  void AdaptInstationary::oneTimestep()
  {
    AMDIS_FUNCNAME("AdaptInstationary::oneTimestep()");

    adaptInfo.setTimestepIteration(0);

    switch (strategy)
    {
    case 0:
      explicitTimeStrategy();
      break;
    case 1:
      implicitTimeStrategy();
      break;
    case 2:
      simpleAdaptiveTimeStrategy();
      break;
    default:
      error_exit("Unknown strategy = {}", strategy);
    }

    adaptInfo.incTimestepNumber();
  }


  int AdaptInstationary::adapt()
  {
    AMDIS_FUNCNAME("AdaptInstationary::adapt()");
    int errorCode = 0;

    test_exit(adaptInfo.getTimestep() >= adaptInfo.getMinTimestep(),
              "timestep < min timestep");
    test_exit(adaptInfo.getTimestep() <= adaptInfo.getMaxTimestep(),
              "timestep > max timestep");

    test_exit(adaptInfo.getTimestep() > 0, "timestep <= 0!");

    if (adaptInfo.getTimestepNumber() == 0)
    {
      adaptInfo.setTime(adaptInfo.getStartTime());
      initialAdaptInfo->setStartTime(adaptInfo.getStartTime());
      initialAdaptInfo->setTime(adaptInfo.getStartTime());

      problemTime->setTime(adaptInfo);

      // initial adaption
      problemTime->solveInitialProblem(*initialAdaptInfo);
      problemTime->transferInitialSolution(adaptInfo);
    }

    while (!adaptInfo.reachedEndTime())
    {
      problemTime->initTimestep(adaptInfo);
      oneTimestep();
      problemTime->closeTimestep(adaptInfo);

      if (breakWhenStable && (adaptInfo.getSolverIterations() == 0))
        break;
    }

    return errorCode;
  }

} // end namespace AMDiS
