#pragma once

#include <list>
#include <memory>

#include <amdis/LocalAssemblerBase.hpp>
#include <amdis/utility/TreeData.hpp>

namespace AMDiS
{
  template <class GridView>
  class OperatorLists
  {
    using Element = typename GridView::template Codim<0>::Entity;
    using Intersection = typename GridView::Intersection;

    template <class OperatorType>
    struct DataElement
    {
      std::shared_ptr<OperatorType> op;
      double* factor = nullptr;
      double* estFactor = nullptr;
      BoundaryType b = {0};
    };

    /// Lists of \ref DataElement on the Element, BoundaryIntersction, and InteriorIntersections
    template <class... Nodes>
    struct Data
    {
      /// The type of local operators associated with grid elements
      using ElementOperator = LocalAssemblerBase<Element, Nodes...>;
      /// The type of local operators associated with grid intersections
      using IntersectionOperator = LocalAssemblerBase<Intersection, Nodes...>;

      /// Return whether any operators are stored on the node
      bool empty() const
      {
        return element.empty() && boundary.empty() && intersection.empty();
      }

      /// Test whether to assemble on the node
      bool doAssemble(bool flag) const
      {
        return flag && (!assembled || changing);
      }

      /// Bind all operators to the grid element and geometry
      template <class Geo>
      void bind(Element const& elem, Geo const& geo)
      {
        for (auto& scaled : element) scaled.op->bind(elem,geo);
        for (auto& scaled : boundary) scaled.op->bind(elem,geo);
        for (auto& scaled : intersection) scaled.op->bind(elem,geo);
      }

      /// Unbind all operators from the element
      void unbind()
      {
        for (auto& scaled : element) scaled.op->unbind();
        for (auto& scaled : boundary) scaled.op->unbind();
        for (auto& scaled : intersection) scaled.op->unbind();
      }

      /// List of operators to be assembled on grid elements
      std::list<DataElement<ElementOperator>> element;
      /// List of operators to be assembled on boundary intersections
      std::list<DataElement<IntersectionOperator>> boundary;
      /// List of operators to be assembled on interior intersections
      std::list<DataElement<IntersectionOperator>> intersection;

      /// if false, do reassemble of all operators
      bool assembled = false;

      /// if true, or \ref assembled false, do reassemble of all operators
      bool changing = false;
    };

  public:

    /// List of operators associated with matrix blocks (RowNode, ColNode)
    template <class RowNode, class ColNode>
    using MatData = Data<RowNode, ColNode>;

    /// List of operators associated with vector blocks [Node]
    template <class Node>
    using VecData = Data<Node>;
  };


  template <class GlobalBasis>
  using MatrixOperators = MatrixData<GlobalBasis, OperatorLists<typename GlobalBasis::GridView>::template MatData>;

  template <class GlobalBasis>
  using VectorOperators = VectorData<GlobalBasis, OperatorLists<typename GlobalBasis::GridView>::template VecData>;

} // end namespace AMDiS
