#pragma once

#include <map>
#include <string>
#include <utility>

#include <dune/common/timer.hh>
#include <dune/typetree/childextraction.hh>

#include <amdis/AdaptInfo.hpp>
#include <amdis/Assembler.hpp>
#include <amdis/FileWriter.hpp>
#include <amdis/LocalAssembler.hpp>
#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/Loops.hpp>
// #include <amdis/Estimator.hpp>

namespace AMDiS {

template <class Traits>
void ProblemStat<Traits>::initialize(
    Flag initFlag,
    Self* adoptProblem,
    Flag adoptFlag)
{
  // create grids
  if (grid_) {
    warning("grid already created");
  }
  else {
    if (initFlag.isSet(CREATE_MESH) ||
        (!adoptFlag.isSet(INIT_MESH) &&
        (initFlag.isSet(INIT_SYSTEM) || initFlag.isSet(INIT_FE_SPACE)))) {
      createGrid();
    }

    if (adoptProblem &&
        (adoptFlag.isSet(INIT_MESH) ||
        adoptFlag.isSet(INIT_SYSTEM) ||
        adoptFlag.isSet(INIT_FE_SPACE))) {
      grid_ = adoptProblem->grid_;
    }
  }

  if (!grid_)
    warning("no grid created");

  int globalRefinements = 0;
  Parameters::get(gridName_ + "->global refinements", globalRefinements);
  if (globalRefinements > 0) {
    grid_->globalRefine(globalRefinements);
  }


  // create fespace
  if (globalBasis_) {
    warning("globalBasis already created");
  }
  else {
    if (initFlag.isSet(INIT_FE_SPACE) ||
        (initFlag.isSet(INIT_SYSTEM) && !adoptFlag.isSet(INIT_FE_SPACE))) {
      createGlobalBasis();
    }

    if (adoptProblem &&
        (adoptFlag.isSet(INIT_FE_SPACE) || adoptFlag.isSet(INIT_SYSTEM))) {
      globalBasis_ = adoptProblem->globalBasis_;
      initGlobalBasis(*globalBasis_);
    }
  }

  if (!globalBasis_)
    warning("no globalBasis created\n");


  // create system
  if (initFlag.isSet(INIT_SYSTEM))
    createMatricesAndVectors();

  if (adoptProblem && adoptFlag.isSet(INIT_SYSTEM)) {
    solution_ = adoptProblem->solution_;
    estimates_ = adoptProblem->estimates_;
    rhs_ = adoptProblem->rhs_;
    systemMatrix_ = adoptProblem->systemMatrix_;
  }


  // create solver
  if (linearSolver_) {
    warning("solver already created\n");
  }
  else {
    if (initFlag.isSet(INIT_SOLVER))
      createSolver();

    if (adoptProblem && adoptFlag.isSet(INIT_SOLVER)) {
      test_exit(!linearSolver_, "solver already created\n");
      linearSolver_ = adoptProblem->linearSolver_;
    }
  }

  if (!linearSolver_) {
    warning("no solver created\n");
  }

  // create marker
    if (initFlag.isSet(INIT_MARKER))
      createMarker();

    if (adoptProblem && adoptFlag.isSet(INIT_MARKER))
      marker_ = adoptProblem->marker_;


  // create file writer
  if (initFlag.isSet(INIT_FILEWRITER))
    createFileWriter();

  solution_->compress();
}


template <class Traits>
void ProblemStat<Traits>::createMarker()
{
  AMDiS::forEachNode_(localView_->tree(), [&,this](auto const& node, auto treePath)
  {
    std::string componentName = name_ + "->marker[" + to_string(treePath) + "]";

    if (!Parameters::get<std::string>(componentName + "->strategy"))
      return;

    std::string tp = to_string(treePath);
    auto newMarker = Marker<Traits>::createMarker(componentName, tp, estimates_[tp], grid_);
    if (newMarker)
      marker_.push_back(std::move(newMarker));

    // If there is more than one marker, and all components are defined
    // on the same grid, the maximum marking has to be enabled.
    // TODO: needs to be checked!
    if (marker_.size() > 1)
      marker_.back()->setMaximumMarking(true);
  });
}


template <class Traits>
void ProblemStat<Traits>::createFileWriter()
{
  forEachNode_(localView_->tree(), [&,this](auto const& node, auto treePath)
  {
    std::string componentName = name_ + "->output[" + to_string(treePath) + "]";

    if (!Parameters::get<std::string>(componentName + "->filename"))
      return;

    auto writer = makeFileWriterPtr(componentName, this->getSolution(treePath));
    filewriter_.push_back(std::move(writer));
  });
}


// add matrix/vector operator terms

template <class Traits>
  template <class Operator, class RowTreePath, class ColTreePath>
void ProblemStat<Traits>::addMatrixOperator(
    Operator const& preOp,
    RowTreePath row, ColTreePath col)
{
  static_assert( Concepts::PreTreePath<RowTreePath>,
      "row must be a valid treepath, or an integer/index-constant");
  static_assert( Concepts::PreTreePath<ColTreePath>,
      "col must be a valid treepath, or an integer/index-constant");

  auto i = child(localView_->tree(), makeTreePath(row));
  auto j = child(localView_->tree(), makeTreePath(col));

  auto op = makeLocalOperator<Element>(preOp, globalBasis_->gridView());
  auto localAssembler = makeLocalAssemblerPtr<Element>(std::move(op), i, j);

  matrixOperators_[i][j].element.push_back({localAssembler, nullptr, nullptr});
  matrixOperators_[i][j].changing = true;
}


template <class Traits>
  template <class Operator, class RowTreePath, class ColTreePath>
void ProblemStat<Traits>::addMatrixOperator(
    BoundaryType b,
    Operator const& preOp,
    RowTreePath row, ColTreePath col)
{
  static_assert( Concepts::PreTreePath<RowTreePath>,
      "row must be a valid treepath, or an integer/index-constant");
  static_assert( Concepts::PreTreePath<ColTreePath>,
      "col must be a valid treepath, or an integer/index-constant");

  auto i = child(localView_->tree(), makeTreePath(row));
  auto j = child(localView_->tree(), makeTreePath(col));
  using Intersection = typename GridView::Intersection;

  auto op = makeLocalOperator<Intersection>(preOp, globalBasis_->gridView());
  auto localAssembler = makeLocalAssemblerPtr<Intersection>(std::move(op), i, j);

  matrixOperators_[i][j].boundary.push_back({localAssembler, nullptr, nullptr, b});
  matrixOperators_[i][j].changing = true;
}


template <class Traits>
  template <class Operator, class TreePath>
void ProblemStat<Traits>::addVectorOperator(
    Operator const& preOp,
    TreePath path)
{
  static_assert( Concepts::PreTreePath<TreePath>,
      "path must be a valid treepath, or an integer/index-constant");

  auto i = child(localView_->tree(), makeTreePath(path));

  auto op = makeLocalOperator<Element>(preOp, globalBasis_->gridView());
  auto localAssembler = makeLocalAssemblerPtr<Element>(std::move(op), i);

  rhsOperators_[i].element.push_back({localAssembler, nullptr, nullptr});
  rhsOperators_[i].changing = true;
}


template <class Traits>
  template <class Operator, class TreePath>
void ProblemStat<Traits>::addVectorOperator(
    BoundaryType b,
    Operator const& preOp,
    TreePath path)
{
  static_assert( Concepts::PreTreePath<TreePath>,
      "path must be a valid treepath, or an integer/index-constant");

  auto i = child(localView_->tree(), makeTreePath(path));
  using Intersection = typename GridView::Intersection;

  auto op = makeLocalOperator<Intersection>(preOp, globalBasis_->gridView());
  auto localAssembler = makeLocalAssemblerPtr<Intersection>(std::move(op), i);

  rhsOperators_[i].boundary.push_back({localAssembler, nullptr, nullptr, b});
  rhsOperators_[i].changing = true;
}


// Adds a Dirichlet boundary condition
template <class Traits>
  template <class Predicate, class RowTreePath, class ColTreePath, class Values>
void ProblemStat<Traits>::
addDirichletBC(Predicate const& predicate, RowTreePath row, ColTreePath col, Values const& values)
{
  static_assert( Concepts::Functor<Predicate, bool(WorldVector)>,
    "Function passed to addDirichletBC for `predicate` does not model the Functor<bool(WorldVector)> concept");

  auto i = child(localView_->tree(), makeTreePath(row));
  auto j = child(localView_->tree(), makeTreePath(col));

  auto valueGridFct = makeGridFunction(values, globalBasis_->gridView());

  using Range = RangeType_t<decltype(i)>;
  using BcType = DirichletBC<WorldVector,Range>;
  auto bc = std::make_shared<BcType>(predicate, valueGridFct);
  constraints_[i][j].push_back(bc);
  // TODO: make DirichletBC an abstract class and add specialization with gridfunction type
}


template <class Traits>
void ProblemStat<Traits>::
solve(AdaptInfo& adaptInfo, bool createMatrixData, bool storeMatrixData)
{
  Dune::Timer t;

  SolverInfo solverInfo(name_ + "->solver");
  solverInfo.setCreateMatrixData(createMatrixData);
  solverInfo.setStoreMatrixData(storeMatrixData);

  solution_->compress();

  linearSolver_->solve(systemMatrix_->matrix(), solution_->vector(), rhs_->vector(),
                      solverInfo);

  if (solverInfo.getInfo() > 0) {
    msg("solution of discrete system needed {} seconds", t.elapsed());

    if (solverInfo.getAbsResidual() >= 0.0) {
      if (solverInfo.getRelResidual() >= 0.0)
        msg("Residual norm: ||b-Ax|| = {}, ||b-Ax||/||b|| = {}",
          solverInfo.getAbsResidual(), solverInfo.getRelResidual());
      else
        msg("Residual norm: ||b-Ax|| = {}", solverInfo.getAbsResidual());
    }
  }

  if (solverInfo.doBreak()) {
    std::stringstream tol_str;
    if (solverInfo.getAbsTolerance() > 0
        && solverInfo.getAbsResidual() > solverInfo.getAbsTolerance())
      tol_str << "absTol = " << solverInfo.getAbsTolerance() << " ";
    if (solverInfo.getRelTolerance() > 0
        && solverInfo.getRelResidual() > solverInfo.getRelTolerance())
      tol_str << "relTol = " << solverInfo.getRelTolerance() << " ";
    error_exit("Tolerance {} could not be reached!", tol_str.str());
  }
}


template <class Traits>
Flag ProblemStat<Traits>::markElements(AdaptInfo& adaptInfo)
{
  Dune::Timer t;

  Flag markFlag = 0;
  for (auto& currentMarker : marker_)
    markFlag |= currentMarker->markGrid(adaptInfo);

  msg("markElements needed {} seconds", t.elapsed());

  return markFlag;
}


template <class Traits>
void ProblemStat<Traits>::
buildAfterCoarsen(AdaptInfo& /*adaptInfo*/, Flag /*flag*/, bool asmMatrix, bool asmVector)
{
  Dune::Timer t;

  Assembler<Traits> assembler(*globalBasis_, matrixOperators_, rhsOperators_, constraints_);
  auto gv = grid_->leafGridView();
  globalBasis_->update(gv);
  assembler.assemble(*systemMatrix_, *solution_, *rhs_, asmMatrix, asmVector);

  msg("buildAfterCoarsen needed {} seconds", t.elapsed());
}


template <class Traits>
void ProblemStat<Traits>::
writeFiles(AdaptInfo& adaptInfo, bool force)
{
  Dune::Timer t;
  for (auto writer : filewriter_)
    writer->writeFiles(adaptInfo, force);
  msg("writeFiles needed {} seconds", t.elapsed());
}


} // end namespace AMDiS
