#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_partialtrial
    {
      std::size_t comp;
    };
  }


  /// first-order operator \f$ \langle\psi, c\,\partial_i\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test_partialtrial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_partialtrial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_partialtrial tag, GridFct const& expr)
      : Super(expr, 1)
      , comp_(tag.comp)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;

      using RangeType = typename RowLocalBasisType::Traits::RangeType;
      using RangeFieldType = typename ColLocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename ColLocalBasisType::Traits::JacobianType;

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);
        assert(jacobian.M() == Context::dim);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto c = Super::coefficient(local);

        thread_local std::vector<RangeType> shapeValues;
        rowLocalFE.localBasis().evaluateFunction(local, shapeValues);

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        colLocalFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<RangeFieldType> colPartial;
        colPartial.resize(shapeGradients.size());
        for (std::size_t i = 0; i < colPartial.size(); ++i) {
          colPartial[i] = jacobian[comp_][0] * shapeGradients[i][0][0];
          for (std::size_t j = 1; j < jacobian.M(); ++j)
            colPartial[i] += jacobian[comp_][j] * shapeGradients[i][0][j];
        }

        for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
          const auto local_j = colNode.localIndex(j);
          const auto value = factor * (c * colPartial[j]);
          for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
            const auto local_i = rowNode.localIndex(i);
            elementMatrix[local_i][local_j] += value * shapeValues[i];
          }
        }
      }
    }

  private:
    std::size_t comp_;
  };

  /** @} **/

} // end namespace AMDiS
