#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec_gradtrial {};
  }


  /// first-order operator \f$ \langle\Psi, c\,\nabla\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec_gradtrial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec_gradtrial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::testvec_gradtrial, GridFct const& expr)
      : Super(expr, 1)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isLeaf,
        "ColNode must be Leaf-Node and RowNode must be a Power-Node.");

      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( CHILDREN == Context::dow, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;

      using RangeType = typename RowLocalBasisType::Traits::RangeType;
      using RangeFieldType = typename ColLocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename ColLocalBasisType::Traits::JacobianType;

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        thread_local std::vector<RangeType> shapeValues;
        rowLocalFE.localBasis().evaluateFunction(local, shapeValues);

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        colLocalFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > colGradients;
        colGradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], colGradients[i]);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value = factor * shapeValues[i];
          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto local_j = colNode.localIndex(j);
            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              elementMatrix[local_ki][local_j] += value * colGradients[j][k];
            }
          }
        }
      }

    }
  };

  /** @} **/

} // end namespace AMDiS
