#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct divtestvec_divtrialvec {};
  }


  /// second-order operator \f$ \langle\nabla\cdot\Psi, c\,\nabla\cdot\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::divtestvec_divtrialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::divtestvec_divtrialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::divtestvec_divtrialvec, GridFct const& expr)
      : Super(expr, 2)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isPower,
        "Operator can be applied to Power-Nodes only.");

      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix);
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix);
    }

  protected:
    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( RowNode::CHILDREN == ColNode::CHILDREN, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;

      using RowFieldType = typename RowLocalBasisType::Traits::RangeFieldType;
      using ColFieldType = typename ColLocalBasisType::Traits::RangeFieldType;
      using RowJacobianType = typename RowLocalBasisType::Traits::JacobianType;
      using ColJacobianType = typename ColLocalBasisType::Traits::JacobianType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        // The gradients of the shape functions on the reference element
        thread_local std::vector<RowJacobianType> rowShapeGradients;
        rowLocalFE.localBasis().evaluateJacobian(local, rowShapeGradients);

        thread_local std::vector<ColJacobianType> colShapeGradients;
        colLocalFE.localBasis().evaluateJacobian(local, colShapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RowFieldType,Context::dow> > rowGradients;
        rowGradients.resize(rowShapeGradients.size());
        for (std::size_t i = 0; i < rowGradients.size(); ++i)
          jacobian.mv(rowShapeGradients[i][0], rowGradients[i]);

        thread_local std::vector<Dune::FieldVector<ColFieldType,Context::dow> > colGradients;
        colGradients.resize(colShapeGradients.size());
        for (std::size_t i = 0; i < colGradients.size(); ++i)
          jacobian.mv(colShapeGradients[i][0], colGradients[i]);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_ki][local_kj] += factor * rowGradients[i][k] * colGradients[j][k];
            }
          }
        }
      }
    }


    template <class Context, class QuadratureRule, class Node, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   Node const& node,
                                   Node const& /*colNode*/,
                                   ElementMatrix& elementMatrix)
    {
      static const std::size_t CHILDREN = Node::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeFieldType = typename LocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename LocalBasisType::Traits::JacobianType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        localFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t k = 0; k < CHILDREN; ++k) {
          for (std::size_t i = 0; i < localFE.size(); ++i) {
            const auto local_ki = node.child(k).localIndex(i);
            const auto value = factor * gradients[i][k];
            elementMatrix[local_ki][local_ki] += value * gradients[i][k];

            for (std::size_t j = i+1; j < localFE.size(); ++j) {
              const auto local_kj = node.child(k).localIndex(j);
              elementMatrix[local_ki][local_kj] += value * gradients[j][k];
              elementMatrix[local_kj][local_ki] += value * gradients[j][k];
            }
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
