#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/Output.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct gradtest_gradtrial {};
  }


  /// second-order operator \f$ \langle\nabla\psi, c\,\nabla\phi\rangle \f$, or \f$ \langle\nabla\psi, A\,\nabla\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::gradtest_gradtrial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::gradtest_gradtrial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    using expr_value_type = typename GridFct::Range;
    static_assert( Category::Scalar<expr_value_type> || Category::Matrix<expr_value_type>,
      "Expression must be of scalar or matrix type." );

  public:
    GridFunctionOperator(tag::gradtest_gradtrial, GridFct const& expr)
      : Super(expr, 2)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();
      using Category = ValueCategory_t<typename GridFct::Range>;

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix, Category{});
      else if (sameFE)
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix);
      else
        error_exit("Not implemented: currently only the implementation for equal fespaces available");
    }

  protected:

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix)
    {
      auto const& localFE = rowNode.finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeFieldType = typename LocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename LocalBasisType::Traits::JacobianType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto exprValue = Super::coefficient(local);

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        localFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto local_i = rowNode.localIndex(i);
          for (std::size_t j = 0; j < localFE.size(); ++j) {
            const auto local_j = colNode.localIndex(j);
            elementMatrix[local_i][local_j] += eval(exprValue, factor, gradients[i], gradients[j]);
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class Node, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   Node const& node,
                                   Node const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::scalar)
    {
      auto const& localFE = node.finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeFieldType = typename LocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename LocalBasisType::Traits::JacobianType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        localFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto local_i = node.localIndex(i);

          elementMatrix[local_i][local_i] += factor * (gradients[i] * gradients[i]);

          for (std::size_t j = i+1; j < localFE.size(); ++j) {
            const auto local_j = node.localIndex(j);
            const auto value = factor * (gradients[i] * gradients[j]);

            elementMatrix[local_i][local_j] += value;
            elementMatrix[local_j][local_i] += value;
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class Node, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   Node const& node,
                                   Node const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::matrix)
    {
      auto const& localFE = node.finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeFieldType = typename LocalBasisType::Traits::RangeFieldType;
      using JacobianType = typename LocalBasisType::Traits::JacobianType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto exprValue = Super::coefficient(local);

        // The gradients of the shape functions on the reference element
        thread_local std::vector<JacobianType> shapeGradients;
        localFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto local_i = node.localIndex(i);
          for (std::size_t j = 0; j < localFE.size(); ++j) {
            const auto local_j = node.localIndex(j);
            elementMatrix[local_i][local_j] += eval(exprValue, factor, gradients[i], gradients[j]);
          }
        }
      }
    }

  protected:

    template <class S, class F, class T, int dow,
      std::enable_if_t<Category::Scalar<S>,int> = 0>
    T eval(S const& scalar, F factor,
           Dune::FieldVector<T,dow> const& grad_test,
           Dune::FieldVector<T,dow> const& grad_trial) const
    {
      return (scalar * factor) * (grad_test * grad_trial);
    }

    template <class M, class F, class T, int dow,
      std::enable_if_t<Category::Matrix<M>,int> = 0>
    T eval(M const& mat, F factor,
           Dune::FieldVector<T,dow> const& grad_test,
           Dune::FieldVector<T,dow> const& grad_trial) const
    {
      return factor * (grad_test * (mat * grad_trial));
    }
  };

  /** @} **/

} // end namespace AMDiS
