#pragma once

#include <type_traits>
#include <vector>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct stokes {};
  }


  /// Stokes operator \f$ \langle\nabla\mathbf{v}, \nu \nabla\mathbf{u}\rangle + \langle\nabla\cdot\mathbf{v}, p\rangle + \langle q, \langle\nabla\cdot\mathbf{u}\rangle \f$
  template <class LocalContext, class ViscosityExpr>
  class GridFunctionOperator<tag::stokes, LocalContext, ViscosityExpr>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::stokes, LocalContext, ViscosityExpr>,
                                        LocalContext, ViscosityExpr>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, ViscosityExpr>;

    static_assert( Category::Scalar<typename ViscosityExpr::Range>, "Viscosity must be of scalar type." );

  public:
    GridFunctionOperator(tag::stokes, ViscosityExpr const& expr)
      : Super(expr, 1)
    {}

    template <class Context, class Node, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          Node const& tree,
                          Node const& /*colNode*/,
                          ElementMatrix& elementMatrix)
    {
      using VelocityNode = typename Node::template Child<0>::type;
      using PressureNode = typename Node::template Child<1>::type;
      static_assert(VelocityNode::isPower && PressureNode::isLeaf, "");

      using namespace Dune::Indices;

      auto const& velocityLocalFE = tree.child(_0,0).finiteElement();
      auto const& pressureLocalFE = tree.child(_1).finiteElement();

      using VelocityLocalBasisType = typename std::decay_t<decltype(velocityLocalFE)>::Traits::LocalBasisType;
      using PressureLocalBasisType = typename std::decay_t<decltype(pressureLocalFE)>::Traits::LocalBasisType;

      using PressureRange = typename PressureLocalBasisType::Traits::RangeType;
      using RangeFieldType = typename VelocityLocalBasisType::Traits::RangeFieldType;
      using VelocityJacobian = typename VelocityLocalBasisType::Traits::JacobianType;

      auto const& quad = this->getQuadratureRule(context.type(), tree, tree);
      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The transposed inverse Jacobian of the map from the reference element to the element
        const auto jacobian = context.geometry().jacobianInverseTransposed(local);

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto vel_factor = Super::coefficient(local) * factor;

        // The gradients of the shape functions on the reference element
        thread_local std::vector<VelocityJacobian> shapeGradients;
        velocityLocalFE.localBasis().evaluateJacobian(local, shapeGradients);

        // Compute the shape function gradients on the real element
        thread_local std::vector<Dune::FieldVector<RangeFieldType,Context::dow> > gradients;
        gradients.resize(shapeGradients.size());
        for (std::size_t i = 0; i < gradients.size(); ++i)
          jacobian.mv(shapeGradients[i][0], gradients[i]);

        thread_local std::vector<PressureRange> pressureValues;
        pressureLocalFE.localBasis().evaluateFunction(local, pressureValues);

        // <viscosity * grad(u), grad(v)>
        for (std::size_t i = 0; i < velocityLocalFE.size(); ++i) {
          const auto value_ii = vel_factor * (gradients[i] * gradients[i]);
          for (std::size_t k = 0; k < Context::dow; ++k) {
            const auto local_ki = tree.child(_0,k).localIndex(i);
            elementMatrix[local_ki][local_ki] += value_ii;
          }

          for (std::size_t j = i+1; j < velocityLocalFE.size(); ++j) {
            const auto value = vel_factor * (gradients[i] * gradients[j]);
            for (std::size_t k = 0; k < Context::dow; ++k) {
              const auto local_ki = tree.child(_0,k).localIndex(i);
              const auto local_kj = tree.child(_0,k).localIndex(j);
              elementMatrix[local_ki][local_kj] += value;
              elementMatrix[local_kj][local_ki] += value;
            }
          }
        }

        // <p, div(v)> + <div(u), q>
        for (std::size_t i = 0; i < velocityLocalFE.size(); ++i) {
          for (std::size_t j = 0; j < pressureLocalFE.size(); ++j) {
            const auto value = factor * pressureValues[j];
            for (std::size_t k = 0; k < Context::dow; ++k) {
              const auto local_v = tree.child(_0,k).localIndex(i);
              const auto local_p = tree.child(_1).localIndex(j);

              elementMatrix[local_v][local_p] += gradients[i][k] * value;
              elementMatrix[local_p][local_v] += gradients[i][k] * value;
            }
          }
        }
      }

    } // end calculateElementMatrix
  };

  /** @} **/

} // end namespace AMDiS
