#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trial {};
  }


  /// zero-order operator \f$ \langle\psi, c\,\phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test_trial, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trial, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Scalar<typename GridFct::Range>, "Expression must be of scalar type." );

  public:
    GridFunctionOperator(tag::test_trial, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix);
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix);
    }

  protected:
    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;
      using RowRangeType = typename RowLocalBasisType::Traits::RangeType;
      using ColRangeType = typename ColLocalBasisType::Traits::RangeType;

      thread_local std::vector<RowRangeType> rowShapeValueBuffer;
      thread_local std::vector<ColRangeType> colShapeValueBuffer;

      const bool sameFE = std::is_same<decltype(rowLocalFE),decltype(colLocalFE)>::value;
      auto* rowShapeValues = &rowShapeValueBuffer;
      auto* colShapeValues = (sameFE ? &rowShapeValueBuffer : &colShapeValueBuffer);

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        rowLocalFE.localBasis().evaluateFunction(local, rowShapeValueBuffer);
        if (!sameFE)
          colLocalFE.localBasis().evaluateFunction(local, colShapeValueBuffer);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto local_i = rowNode.localIndex(i);
          const auto value = factor * (*rowShapeValues)[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto local_j = colNode.localIndex(j);
            elementMatrix[local_i][local_j] += value * (*colShapeValues)[j];
          }
        }
      }
    }


    template <class Context, class QuadratureRule,
              class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isLeaf,
        "Operator can be applied to Leaf-Nodes only.");

      auto const& localFE = node.finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeType = typename LocalBasisType::Traits::RangeType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        thread_local std::vector<RangeType> shapeValues;
        localFE.localBasis().evaluateFunction(local, shapeValues);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto local_i = node.localIndex(i);

          const auto value = factor * shapeValues[i];
          elementMatrix[local_i][local_i] += value * shapeValues[i];

          for (std::size_t j = i+1; j < localFE.size(); ++j) {
            const auto local_j = node.localIndex(j);

            elementMatrix[local_i][local_j] += value * shapeValues[j];
            elementMatrix[local_j][local_i] += value * shapeValues[j];
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
