#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct test_trialvec {};
  }


  /// zero-order operator \f$ \langle\psi, \mathbf{b}\cdot\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::test_trialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::test_trialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Vector<typename GridFct::Range>, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::test_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isLeaf && ColNode::isPower,
        "RowNode must be Leaf-Node and ColNode must be a Power-Node.");

      static const std::size_t CHILDREN = ColNode::CHILDREN;
      static_assert( std::is_same<typename GridFct::Range, Dune::FieldVector<double, CHILDREN>>::value, "" );

      auto const& rowLocalFE = rowNode.finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;
      using RowRangeType = typename RowLocalBasisType::Traits::RangeType;
      using ColRangeType = typename ColLocalBasisType::Traits::RangeType;

      thread_local std::vector<RowRangeType> rowShapeValueBuffer;
      thread_local std::vector<ColRangeType> colShapeValueBuffer;

      const bool sameFE = std::is_same<decltype(rowLocalFE),decltype(colLocalFE)>::value;
      auto* rowShapeValues = &rowShapeValueBuffer;
      auto* colShapeValues = (sameFE ? &rowShapeValueBuffer : &colShapeValueBuffer);

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto b = Super::coefficient(local);

        rowLocalFE.localBasis().evaluateFunction(local, rowShapeValueBuffer);
        if (!sameFE)
          colLocalFE.localBasis().evaluateFunction(local, colShapeValueBuffer);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto local_i = rowNode.localIndex(i);

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value = b * (factor * (*rowShapeValues)[i] * (*colShapeValues)[j]);

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_kj = colNode.child(k).localIndex(j);
              elementMatrix[local_i][local_kj] += value[k];
            }
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
