#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec {};
  }


  /// zero-order vector-operator \f$ (\mathbf{b}\cdot\Psi) \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    static_assert( Category::Vector<typename GridFct::Range>, "Expression must be of vector type." );

  public:
    GridFunctionOperator(tag::testvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class Node, class ElementVector>
    void getElementVector(Context const& context,
                          Node const& node,
                          ElementVector& elementVector)
    {
      static_assert(Node::isPower,
        "Operator can be applied to Power-Nodes only.");

      static const std::size_t CHILDREN = Node::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeType = typename LocalBasisType::Traits::RangeType;

      auto const& quad = this->getQuadratureRule(context.type(), node);
      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const auto exprValue = Super::coefficient(local);

        thread_local std::vector<RangeType> shapeValues;
        localFE.localBasis().evaluateFunction(local, shapeValues);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto value = exprValue * (factor * shapeValues[i]);
          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_ki = node.child(k).localIndex(i);
            elementVector[local_ki] += value[k];
          }
        }
      }
    }
  };

  /** @} **/

} // end namespace AMDiS
