#pragma once

#include <type_traits>

#include <amdis/GridFunctionOperator.hpp>
#include <amdis/common/ValueCategory.hpp>

namespace AMDiS
{
  /**
   * \addtogroup operators
   * @{
   **/

  namespace tag
  {
    struct testvec_trialvec {};
  }


  /// zero-order operator \f$ \langle\Psi, c\,\Phi\rangle \f$, or \f$ \langle\Psi, A\,\Phi\rangle \f$
  template <class LocalContext, class GridFct>
  class GridFunctionOperator<tag::testvec_trialvec, LocalContext, GridFct>
      : public GridFunctionOperatorBase<GridFunctionOperator<tag::testvec_trialvec, LocalContext, GridFct>,
                                        LocalContext, GridFct>
  {
    using Self = GridFunctionOperator;
    using Super = GridFunctionOperatorBase<Self, LocalContext, GridFct>;

    using expr_value_type = typename GridFct::Range;
    static_assert( Category::Scalar<expr_value_type> || Category::Matrix<expr_value_type>,
      "Expression must be of scalar or matrix type." );

  public:
    GridFunctionOperator(tag::testvec_trialvec, GridFct const& expr)
      : Super(expr, 0)
    {}

    template <class Context, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrix(Context const& context,
                          RowNode const& rowNode,
                          ColNode const& colNode,
                          ElementMatrix& elementMatrix)
    {
      static_assert(RowNode::isPower && ColNode::isPower,
        "Operator can be applied to Power-Nodes only.");

      static_assert( (RowNode::CHILDREN == ColNode::CHILDREN), "" );
      // theoretically the restriction of the same nr of childs would not be necessary

      const bool sameFE = std::is_same<FiniteElementType_t<RowNode>, FiniteElementType_t<ColNode>>::value;
      const bool sameNode = rowNode.treeIndex() == colNode.treeIndex();

      using Category = ValueCategory_t<expr_value_type>;

      auto const& quad = this->getQuadratureRule(context.type(), rowNode, colNode);
      if (sameFE && sameNode && std::is_same<Category,tag::scalar>::value)
        getElementMatrixOptimized(context, quad, rowNode, colNode, elementMatrix, Category{});
      else
        getElementMatrixStandard(context, quad, rowNode, colNode, elementMatrix, Category{});
    }

  protected: // specialization for different value-categories

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix,
                                  tag::scalar)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;
      using RowRangeType = typename RowLocalBasisType::Traits::RangeType;
      using ColRangeType = typename ColLocalBasisType::Traits::RangeType;

      thread_local std::vector<RowRangeType> rowShapeValueBuffer;
      thread_local std::vector<ColRangeType> colShapeValueBuffer;

      const bool sameFE = std::is_same<decltype(rowLocalFE),decltype(colLocalFE)>::value;
      auto* rowShapeValues = &rowShapeValueBuffer;
      auto* colShapeValues = (sameFE ? &rowShapeValueBuffer : &colShapeValueBuffer);

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        rowLocalFE.localBasis().evaluateFunction(local, rowShapeValueBuffer);
        if (!sameFE)
          colLocalFE.localBasis().evaluateFunction(local, colShapeValueBuffer);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value = factor * (*rowShapeValues)[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value0 = value * (*colShapeValues)[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = rowNode.child(k).localIndex(i);
              const auto local_kj = colNode.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
            }
          }
        }
      }
    }


    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::scalar)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;

      auto const& localFE = node.child(0).finiteElement();

      using LocalBasisType = typename std::decay_t<decltype(localFE)>::Traits::LocalBasisType;
      using RangeType = typename LocalBasisType::Traits::RangeType;

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = Super::coefficient(local) * context.integrationElement(qp.position()) * qp.weight();

        thread_local std::vector<RangeType> shapeValues;
        localFE.localBasis().evaluateFunction(local, shapeValues);

        for (std::size_t i = 0; i < localFE.size(); ++i) {
          const auto value = factor * shapeValues[i];

          for (std::size_t k = 0; k < CHILDREN; ++k) {
            const auto local_ki = node.child(k).localIndex(i);
            elementMatrix[local_ki][local_ki] += value * shapeValues[i];
          }

          for (std::size_t j = i+1; j < localFE.size(); ++j) {
            const auto value0 = value * shapeValues[j];

            for (std::size_t k = 0; k < CHILDREN; ++k) {
              const auto local_ki = node.child(k).localIndex(i);
              const auto local_kj = node.child(k).localIndex(j);

              elementMatrix[local_ki][local_kj] += value0;
              elementMatrix[local_kj][local_ki] += value0;
            }
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixStandard(Context const& context,
                                  QuadratureRule const& quad,
                                  RowNode const& rowNode,
                                  ColNode const& colNode,
                                  ElementMatrix& elementMatrix,
                                  tag::matrix)
    {
      static const std::size_t CHILDREN = RowNode::CHILDREN;
      static_assert( std::is_same<expr_value_type, Dune::FieldMatrix<double, CHILDREN, CHILDREN>>::value, "" );

      auto const& rowLocalFE = rowNode.child(0).finiteElement();
      auto const& colLocalFE = colNode.child(0).finiteElement();

      using RowLocalBasisType = typename std::decay_t<decltype(rowLocalFE)>::Traits::LocalBasisType;
      using ColLocalBasisType = typename std::decay_t<decltype(colLocalFE)>::Traits::LocalBasisType;
      using RowRangeType = typename RowLocalBasisType::Traits::RangeType;
      using ColRangeType = typename ColLocalBasisType::Traits::RangeType;

      thread_local std::vector<RowRangeType> rowShapeValueBuffer;
      thread_local std::vector<ColRangeType> colShapeValueBuffer;

      const bool sameFE = std::is_same<decltype(rowLocalFE),decltype(colLocalFE)>::value;
      auto* rowShapeValues = &rowShapeValueBuffer;
      auto* colShapeValues = (sameFE ? &rowShapeValueBuffer : &colShapeValueBuffer);

      for (auto const& qp : quad) {
        // Position of the current quadrature point in the reference element
        decltype(auto) local = context.local(qp.position());

        // The multiplicative factor in the integral transformation formula
        const auto factor = context.integrationElement(qp.position()) * qp.weight();
        const Dune::FieldMatrix<double, CHILDREN, CHILDREN> exprValue = Super::coefficient(local);

        rowLocalFE.localBasis().evaluateFunction(local, rowShapeValueBuffer);
        if (!sameFE)
          colLocalFE.localBasis().evaluateFunction(local, colShapeValueBuffer);

        for (std::size_t i = 0; i < rowLocalFE.size(); ++i) {
          const auto value0 = factor * (*rowShapeValues)[i];

          for (std::size_t j = 0; j < colLocalFE.size(); ++j) {
            const auto value = value0 * (*colShapeValues)[j];
            const auto mat = exprValue * value;

            for (std::size_t k0 = 0; k0 < CHILDREN; ++k0) {
              const auto local_ki = rowNode.child(k0).localIndex(i);
              for (std::size_t k1 = 0; k1 < CHILDREN; ++k1) {
                const auto local_kj = colNode.child(k1).localIndex(j);

                elementMatrix[local_ki][local_kj] += mat[k0][k1];
              }
            }
          }
        }
      }
    }

    template <class Context, class QuadratureRule, class RowNode, class ColNode, class ElementMatrix>
    void getElementMatrixOptimized(Context const& context,
                                   QuadratureRule const& quad,
                                   RowNode const& node,
                                   ColNode const& /*colNode*/,
                                   ElementMatrix& elementMatrix,
                                   tag::matrix)
    {
      assert(false && "Not implemented.");
    }
  };

  /** @} **/

} // end namespace AMDiS
