#pragma once

#include <dune/common/diagonalmatrix.hh>
#include <dune/common/fmatrix.hh>
#include <dune/common/fvector.hh>

#include <amdis/common/ConceptsBase.hpp>
#include <amdis/common/ScalarTypes.hpp>

namespace AMDiS
{
  using Dune::FieldVector;
  using Dune::FieldMatrix;

  // some arithmetic operations with FieldVector

  template <class T, int N, class S,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldVector<T,N> operator*(FieldVector<T,N> v, S factor);

  template <class S, class T, int N,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldVector<T,N> operator*(S factor, FieldVector<T,N> v);

  template <class T, int N, class S,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldVector<T,N> operator/(FieldVector<T,N> v, S factor);

  template <class T>
  FieldVector<T,1> operator*(FieldVector<T,1> const& v, FieldVector<T,1> const& w);

  template <class T, int N>
  FieldVector<T,N> operator*(FieldVector<T,1> const& factor, FieldVector<T,N> v);

  template <class T, int N>
  FieldVector<T,N> operator*(FieldVector<T,N> v, FieldVector<T,1> const& factor);

  // ----------------------------------------------------------------------------

  /// Cross-product a 2d-vector = orthogonal vector
  template <class T>
  FieldVector<T, 2> cross(FieldVector<T, 2> const& a);

  /// Cross-product of two 3d-vectors
  template <class T>
  FieldVector<T, 3> cross(FieldVector<T, 3> const& a, FieldVector<T, 3> const& b);

  /// Dot product (vec1^T * vec2)
  template <class T, class S, int N>
  auto dot(FieldVector<T,N> const& vec1, FieldVector<S,N> const& vec2);

  template <class T, int N, int M,
    REQUIRES( N!=1 && M!=1 )>
  auto operator*(FieldVector<T,N> const& v, FieldVector<T,M> const& w);

  template <class T, class S, int N>
  auto dot(FieldMatrix<T,1,N> const& vec1, FieldMatrix<S,1,N> const& vec2);

  // ----------------------------------------------------------------------------

  /// Sum of vector entires.
  template <class T, int N>
  T sum(FieldVector<T, N> const& x);

  template <class T, int N>
  T sum(FieldMatrix<T, 1, N> const& x);


  /// Dot-product with the vector itself
  template <class T, int N>
  auto unary_dot(FieldVector<T, N> const& x);

  template <class T, int N>
  auto unary_dot(FieldMatrix<T, 1, N> const& x);

  /// Maximum over all vector entries
  template <class T, int N>
  auto max(FieldVector<T, N> const& x);

  template <class T, int N>
  auto max(FieldMatrix<T, 1, N> const& x);

  /// Minimum over all vector entries
  template <class T, int N>
  auto min(FieldVector<T, N> const& x);

  template <class T, int N>
  auto min(FieldMatrix<T, 1, N> const& x);

  /// Maximum of the absolute values of vector entries
  template <class T, int N>
  auto abs_max(FieldVector<T, N> const& x);

  template <class T, int N>
  auto abs_max(FieldMatrix<T, 1, N> const& x);

  /// Minimum of the absolute values of vector entries
  template <class T, int N>
  auto abs_min(FieldVector<T, N> const& x);

  template <class T, int N>
  auto abs_min(FieldMatrix<T, 1, N> const& x);

  // ----------------------------------------------------------------------------

  /** \ingroup vector_norms
   *  \brief The 1-norm of a vector = sum_i |x_i|
   **/
  template <class T, int N>
  auto one_norm(FieldVector<T, N> const& x);

  template <class T, int N>
  auto one_norm(FieldMatrix<T, 1, N> const& x);

  /** \ingroup vector_norms
   *  \brief The euklidean 2-norm of a vector = sqrt( sum_i |x_i|^2 )
   **/
  template <class T, int N>
  auto two_norm(FieldVector<T, N> const& x);

  template <class T, int N>
  auto two_norm(FieldMatrix<T, 1, N> const& x);

  /** \ingroup vector_norms
   *  \brief The p-norm of a vector = ( sum_i |x_i|^p )^(1/p)
   **/
  template <int p, class T, int N>
  auto p_norm(FieldVector<T, N> const& x);

  template <int p, class T, int N>
  auto p_norm(FieldMatrix<T, 1, N> const& x);

  /** \ingroup vector_norms
   *  \brief The infty-norm of a vector = max_i |x_i| = alias for \ref abs_max
   **/
  template <class T, int N>
  auto infty_norm(FieldVector<T, N> const& x);

  template <class T, int N>
  auto infty_norm(FieldMatrix<T, 1, N> const& x);

  // ----------------------------------------------------------------------------

  /// The euklidean distance between two vectors = |lhs-rhs|_2
  template <class T, int N>
  T distance(FieldVector<T, N> const& lhs, FieldVector<T, N> const& rhs);

  // ----------------------------------------------------------------------------

  /// Outer product (vec1 * vec2^T)
  template <class T, class S, int N, int M, int K>
  auto outer(FieldMatrix<T,N,K> const& vec1, FieldMatrix<S,M,K> const& vec2);

  // ----------------------------------------------------------------------------

  template <class T>
  T det(FieldMatrix<T, 0, 0> const& /*mat*/);

  /// Determinant of a 1x1 matrix
  template <class T>
  T det(FieldMatrix<T, 1, 1> const& mat);

  /// Determinant of a 2x2 matrix
  template <class T>
  T det(FieldMatrix<T, 2, 2> const& mat);

  /// Determinant of a 3x3 matrix
  template <class T>
  T det(FieldMatrix<T, 3, 3> const& mat);

  /// Determinant of a NxN matrix
  template <class T,  int N>
  T det(FieldMatrix<T, N, N> const& mat);


  /// Return the inverse of the matrix `mat`
  template <class T, int N>
  auto inv(FieldMatrix<T, N, N> mat);

  /// Solve the linear system A*x = b
  template <class T, int N>
  void solve(FieldMatrix<T, N, N> const& A,  FieldVector<T, N>& x,  FieldVector<T, N> const& b);


  /// Gramian determinant = sqrt( det( DT^T * DF ) )
  template <class T, int N, int M>
  T gramian(FieldMatrix<T,N,M> const& DF);

  /// Gramian determinant, specialization for 1 column matrices
  template <class T, int M>
  T gramian(FieldMatrix<T, 1, M> const& DF);

  // ----------------------------------------------------------------------------
  // some arithmetic operations with FieldMatrix

  template <class T, int M, int N>
  FieldMatrix<T,N,M> trans(FieldMatrix<T, M, N> const& A);


  template <class T, int M, int N, class S,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldMatrix<T,M,N> operator*(S scalar, FieldMatrix<T, M, N> A);

  template <class T, int M, int N, class S,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldMatrix<T,M,N> operator*(FieldMatrix<T, M, N> A, S scalar);

  template <class T, int M, int N, class S,
    REQUIRES(Concepts::Arithmetic<S>) >
  FieldMatrix<T,M,N> operator/(FieldMatrix<T, M, N> A, S scalar);


  template <class T, int M, int N>
  FieldMatrix<T,M,N> operator+(FieldMatrix<T, M, N> A, FieldMatrix<T, M, N> const& B);

  template <class T, int M, int N>
  FieldMatrix<T,M,N> operator-(FieldMatrix<T, M, N> A, FieldMatrix<T, M, N> const& B);


  template <class T, int N, int M>
  FieldVector<T,N> operator*(FieldMatrix<T,N,M> const& mat, FieldVector<T,M> const& vec);

  template <class T, int N, int M>
  FieldMatrix<T,N,M> operator*(FieldMatrix<T,N,M> mat, FieldVector<T,1> const& scalar);

  template <class T, int N>
  FieldMatrix<T,N,1> operator*(FieldMatrix<T,N,1> mat, FieldVector<T,1> const& scalar);



  template <class T, int M, int N, int L>
  FieldMatrix<T,M,N> multiplies(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, L, N> const& B);



  template <class T, int M, int N, int L>
  FieldMatrix<T,M,N> multiplies_AtB(FieldMatrix<T, L, M> const& A,  FieldMatrix<T, N, L> const& B);

  template <class T, int M, int N, int L>
  FieldMatrix<T,M,N> multiplies_ABt(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, N, L> const& B);

  template <class T, int M, int N, int L>
  FieldMatrix<T,M,N>& multiplies_ABt(FieldMatrix<T, M, L> const& A,  FieldMatrix<T, N, L> const& B, FieldMatrix<T,M,N>& C);

  template <class T, int M, int N>
  FieldMatrix<T,M,N>& multiplies_ABt(FieldMatrix<T, M, N> const& A,  Dune::DiagonalMatrix<T, N> const& B, FieldMatrix<T,M,N>& C);

} // end namespace AMDiS

#include "FieldMatVec.inc.hpp"
