#pragma once

#include <cmath>
#include <amdis/common/Math.hpp>

namespace AMDiS
{
  template <class RowBasisType, class ColBasisType, class Backend>
  class DOFMatrixBase
  {
  public:
    /// The type of the finite element space / basis of the row
    using RowBasis = RowBasisType;

    /// The type of the finite element space / basis of the column
    using ColBasis = ColBasisType;

    /// The index/size - type
    using size_type = typename RowBasis::size_type;

    using BaseMatrix = typename Backend::BaseMatrix;

  public:
    /// Constructor. Constructs new BaseVector.
    DOFMatrixBase(RowBasis const& rowBasis, ColBasis const& colBasis)
      : rowBasis_(&rowBasis)
      , colBasis_(&colBasis)
    {}

    /// Return the row-basis \ref rowBasis of the matrix
    RowBasis const& rowBasis() const
    {
      return *rowBasis_;
    }

    /// Return the col-basis \ref colBasis of the matrix
    ColBasis const& colBasis() const
    {
      return *colBasis_;
    }

    /// Return the data-matrix
    BaseMatrix const& matrix() const
    {
      return backend_.matrix();
    }

    /// Return the data-matrix
    BaseMatrix& matrix()
    {
      return backend_.matrix();
    }

    /// Return the size of the \ref rowBasis_
    size_type rows() const
    {
      return rowBasis_->dimension();
    }

    /// Return the size of the \ref colBasis_
    size_type cols() const
    {
      return colBasis_->dimension();
    }

    /// Initialize the matrix for insertion, e.g. allocate the non-zero pattern
    /// If \p setToZero is true, the matrix is set to 0
    void init(bool setToZero)
    {
      backend_.init(*rowBasis_, *colBasis_, setToZero);
    }

    /// Finish the matrix insertion, e.g. cleanup or final insertion
    void finish()
    {
      backend_.finish();
    }

    /// Insert a block of values into the matrix (add to existing values)
    template <class ElementMatrix>
    void insert(typename RowBasis::LocalView const& rowLocalView,
                typename ColBasis::LocalView const& colLocalView,
                ElementMatrix const& elementMatrix)
    {
      for (size_type i = 0; i < rowLocalView.size(); ++i) {
        size_type const row = flatMultiIndex(rowLocalView.index(i));
        for (size_type j = 0; j < colLocalView.size(); ++j) {
          if (std::abs(elementMatrix[i][j]) > threshold<double>) {
            size_type const col = flatMultiIndex(colLocalView.index(j));
            backend_.insert(row, col, elementMatrix[i][j]);
          }
        }
      }
    }

    /// Insert a single value into the matrix (add to existing value)
    template <class RowIndex, class ColIndex>
    void insert(RowIndex row, ColIndex col, typename Backend::value_type const& value)
    {
      backend_.insert(flatMultiIndex(row), flatMultiIndex(col), value);
    }

    /// \brief Deletes all rows with \p dirichletNodes != 0 and if \p apply is true, adds
    /// a one on the diagonal of the matrix.
    auto applyDirichletBC(std::vector<char> const& dirichletNodes)
    {
      return backend_.applyDirichletBC(dirichletNodes);
    }

    size_type nnz() const
    {
      return backend_.nnz();
    }

  protected:
    RowBasis const*	rowBasis_;
    ColBasis const*	colBasis_;

    Backend backend_;
  };

} // end namespace AMDiS
