#pragma once

#include <cmath>

#include <dune/functions/functionspacebases/sizeinfo.hh>

#include <amdis/common/Math.hpp>
#include <amdis/common/ScalarTypes.hpp>
#include <amdis/linear_algebra/DOFVectorInterface.hpp>
#include <amdis/utility/MultiIndex.hpp>

namespace AMDiS
{
  /// The basic container that stores a base vector and a corresponding basis
  template <class BasisType, class Backend>
  class DOFVectorBase
      : public DOFVectorInterface
  {
    using Self = DOFVectorBase;

  public:
    /// The type of the \ref basis
    using Basis = BasisType;

    /// The index/size - type
    using size_type  = typename BasisType::size_type;

    /// The type of the elements of the DOFVector
    using value_type = typename Backend::value_type;

    using BaseVector = typename Backend::BaseVector;

  public:
    /// Constructor. Constructs new BaseVector.
    DOFVectorBase(BasisType const& basis)
      : basis_(&basis)
    {
      compress();
    }

    /// Return the basis \ref basis_ associated to the vector
    Basis const& basis() const
    {
      return *basis_;
    }

    /// Return the data-vector
    BaseVector const& vector() const
    {
      return backend_.vector();
    }

    /// Return the data-vector
    BaseVector& vector()
    {
      return backend_.vector();
    }

    /// Return the size of the \ref basis
    size_type size() const
    {
      return basis_->dimension();
    }

    void resize(Dune::Functions::SizeInfo<Basis> const& s)
    {
      backend_.resize(size_type(s));
    }

    /// Resize the \ref vector to the size of the \ref basis.
    virtual void compress() override
    {
      if (backend_.size() != size()) {
        backend_.resize(size());
        vector() = 0;
      }
    }

    /// Access the entry \p idx of the \ref vector with read-access.
    template <class Index>
    auto const& operator[](Index idx) const
    {
      size_type i = flatMultiIndex(idx);
      return backend_[i];
    }

    /// Access the entry \p idx of the \ref vector with write-access.
    template <class Index>
    auto& operator[](Index idx)
    {
      size_type i = flatMultiIndex(idx);
      return backend_[i];
    }

    /// Insert a block of values into the matrix (add to existing values)
    template <class ElementVector>
    void insert(typename Basis::LocalView const& localView,
                ElementVector const& elementVector)
    {
      for (size_type i = 0; i < localView.size(); ++i) {
        if (std::abs(elementVector[i]) > threshold<double>) {
          size_type const idx = flatMultiIndex(localView.index(i));
          backend_[idx] += elementVector[i];
        }
      }
    }

    /// Sets each DOFVector to the scalar \p value.
    template <class Scalar,
      std::enable_if_t<Concepts::Arithmetic<Scalar>, int> = 0>
    Self& operator=(Scalar value)
    {
      vector() = value;
      return *this;
    }

  private:
    /// The finite element space / basis associated with the data vector
    Basis const* basis_;

    Backend backend_;
  };

} // end namespace AMDiS
